Option Explicit On

Public Class MatrixLibVb

  Private Const DEF_BAD_VAL = -1.1E+100
  Private Shared mfBadDataVal As Double
  Private Shared mnErrCount As Integer
  Private Shared msErrMessage() As String
  Private Shared msErrSrc() As String
  Private Shared m_lPower2(0 To 31) As Long

  Private Shared Sub INC(ByRef N As Integer, Optional ByVal I As Integer = 1)
    N = N + I
  End Sub

  Private Shared Sub DEC(ByRef N As Integer, Optional ByVal I As Integer = 1)
    N = N - I
  End Sub

  Private Shared Sub INCL(ByRef N As Long, Optional ByVal I As Long = 1)
    N = N + I
  End Sub

  Private Shared Sub DECL(ByRef N As Long, Optional ByVal I As Long = 1)
    N = N - I
  End Sub

  Private Shared Sub INCX(ByRef N As Double, Optional ByVal I As Double = 1)
    N = N + I
  End Sub

  Private Shared Sub DECX(ByRef N As Double, Optional ByVal I As Double = 1)
    N = N - I
  End Sub

  Public Shared Sub ResetErrors()

    mnErrCount = 0
    ReDim msErrMessage(mnErrCount)
    ReDim msErrSrc(mnErrCount)
  End Sub

  Public Shared Function GetErrorCount() As Integer
    GetErrorCount = mnErrCount
  End Function


  Private Shared Sub ReportError(ByVal sMessage As String, ByVal sSource As String)

    ' adjust arrays
    ReDim Preserve msErrMessage(mnErrCount)
    ReDim Preserve msErrSrc(mnErrCount)
    msErrMessage(mnErrCount) = sMessage
    msErrSrc(mnErrCount) = sSource
    INC(mnErrCount)
  End Sub


  Public Shared Function GetRecentMessageAndSource( _
          ByRef sErrMsg As String, _
          ByRef sErrSrc As String, _
          ByVal bDecrementErrMessageIndex As Boolean) As Boolean

    ' set default Exit Sub values
    sErrMsg = ""
    sErrSrc = ""
    If mnErrCount > 0 Then
      sErrMsg = msErrMessage(mnErrCount - 1)
      sErrSrc = msErrSrc(mnErrCount - 1)
      If bDecrementErrMessageIndex And mnErrCount > 1 Then
        DEC(mnErrCount)
      End If
      GetRecentMessageAndSource = True
    Else
      GetRecentMessageAndSource = False
    End If
  End Function

  Public Shared Sub GetErrorMessages(ByRef sMsgArr() As String)
    Dim I As Integer

    ReDim sMsgArr(UBound(msErrMessage))
    For I = 0 To mnErrCount - 1
      sMsgArr(I) = msErrMessage(I)
    Next I
  End Sub

  Public Shared Sub GetErrorSources(ByRef sMsgSrc() As String)
    Dim I As Integer

    ReDim sMsgSrc(UBound(msErrSrc))
    For I = 0 To mnErrCount - 1
      sMsgSrc(I) = msErrSrc(I)
    Next I
  End Sub

  Public Shared Function GetBadDataVal() As Double
    GetBadDataVal = mfBadDataVal
  End Function


  Public Shared Function SetBadDataVal(ByVal X As Double) As Double
    Dim Y As Double

    Y = mfBadDataVal
    mfBadDataVal = X
    SetBadDataVal = Y
  End Function

  Private Shared Function Max(ByVal X As Double, ByVal Y As Double) As Double
    Return IIf(X > Y, X, Y)
  End Function

  Private Shared Function Square(ByVal X As Double) As Double
    Return X * X
  End Function

  Private Shared Function Sign(ByVal A As Double, ByVal B As Double) As Double
    Return IIf(B >= 0, IIf(A >= 0, A, -A), IIf(A >= 0, -A, A))
  End Function

  Private Shared Sub Swap(ByRef X As Double, ByRef Y As Double)
    Dim temp As Double

    temp = X
    X = Y
    Y = temp
  End Sub

  Public Shared Function RShift(ByVal lThis As Long, ByVal lBits As Long) As Long
    If (lBits <= 0) Then
      RShift = lThis
    ElseIf (lBits > 63) Then
      ' .. error ...
    ElseIf (lBits > 31) Then
      RShift = 0
    Else
      If (lThis And m_lPower2(31 - lBits)) = m_lPower2(31 - lBits) Then
        RShift = (lThis And (m_lPower2(31 - lBits) - 1)) * m_lPower2(lBits) Or m_lPower2(31)
      Else
        RShift = (lThis And (m_lPower2(31 - lBits) - 1)) * m_lPower2(lBits)
      End If
    End If
  End Function

  Public Shared Function LShift(ByVal lThis As Long, ByVal lBits As Long) As Long

    If (lBits <= 0) Then
      LShift = lThis
    ElseIf (lBits > 63) Then
      ' ... error ...
    ElseIf (lBits > 31) Then
      LShift = 0
    Else
      If (lThis And m_lPower2(31)) = m_lPower2(31) Then
        LShift = (lThis And &H7FFFFFFF) \ m_lPower2(lBits) Or m_lPower2(31 - lBits)
      Else
        LShift = lThis \ m_lPower2(lBits)
      End If
    End If
  End Function

  Public Shared Sub InitShiftData()
    m_lPower2(0) = &H1&
    m_lPower2(1) = &H2&
    m_lPower2(2) = &H4&
    m_lPower2(3) = &H8&
    m_lPower2(4) = &H10&
    m_lPower2(5) = &H20&
    m_lPower2(6) = &H40&
    m_lPower2(7) = &H80&
    m_lPower2(8) = &H100&
    m_lPower2(9) = &H200&
    m_lPower2(10) = &H400&
    m_lPower2(11) = &H800&
    m_lPower2(12) = &H1000&
    m_lPower2(13) = &H2000&
    m_lPower2(14) = &H4000&
    m_lPower2(15) = &H8000&
    m_lPower2(16) = &H10000
    m_lPower2(17) = &H20000
    m_lPower2(18) = &H40000
    m_lPower2(19) = &H80000
    m_lPower2(20) = &H100000
    m_lPower2(21) = &H200000
    m_lPower2(22) = &H400000
    m_lPower2(23) = &H800000
    m_lPower2(24) = &H1000000
    m_lPower2(25) = &H2000000
    m_lPower2(26) = &H4000000
    m_lPower2(27) = &H8000000
    m_lPower2(28) = &H10000000
    m_lPower2(29) = &H20000000
    m_lPower2(30) = &H40000000
    m_lPower2(31) = &H80000000
  End Sub


  Public Shared Function isFullOfEqualTo(ByVal A(,) As Double, ByVal fValue As Double) As Boolean

    Dim nRow As Integer
    Dim nCol As Integer

    isFullOfEqualTo = False
    For nRow = 0 To UBound(A, 1)
      For nCol = 0 To UBound(A, 2)
        If fValue <> A(nRow, nCol) Then Exit Function
      Next nCol
    Next nRow
    Return True
  End Function

  Public Shared Function isFullOfEqOrGreaterThan(ByVal A(,) As Double, ByVal fValue As Double) As Boolean

    Dim nRow As Integer
    Dim nCol As Integer

    isFullOfEqOrGreaterThan = False
    For nRow = 0 To UBound(A, 1)
      For nCol = 0 To UBound(A, 2)
        If A(nRow, nCol) < fValue Then Exit Function
      Next nCol
    Next nRow
    isFullOfEqOrGreaterThan = True
  End Function


  Public Shared Function isFullOfGreaterThan(ByVal A(,) As Double, ByVal fValue As Double) As Boolean

    Dim nRow As Integer
    Dim nCol As Integer

    isFullOfGreaterThan = False
    For nRow = 0 To UBound(A, 1)
      For nCol = 0 To UBound(A, 2)
        If A(nRow, nCol) <= fValue Then Exit Function
      Next nCol
    Next nRow
    isFullOfGreaterThan = True
  End Function


  Public Shared Function isFullOfEqOrLessThan(ByVal A(,) As Double, ByVal fValue As Double) As Boolean

    Dim nRow As Integer
    Dim nCol As Integer

    isFullOfEqOrLessThan = False
    For nRow = 0 To UBound(A, 1)
      For nCol = 0 To UBound(A, 2)
        If A(nRow, nCol) > fValue Then Exit Function
      Next nCol
    Next nRow
    isFullOfEqOrLessThan = True
  End Function


  Public Shared Function isFullOfLess(ByVal A(,) As Double, ByVal fValue As Double) As Boolean
    Dim nRow As Integer
    Dim nCol As Integer

    isFullOfLess = False
    For nRow = 0 To UBound(A, 1)
      For nCol = 0 To UBound(A, 2)
        If A(nRow, nCol) >= fValue Then Exit Function
      Next nCol
    Next nRow
    isFullOfLess = True
  End Function

  Public Shared Sub MatchMatDim(ByRef SourceMat(,) As Double, ByRef TargetMat(,) As Double)
    ReDim TargetMat(UBound(SourceMat, 1), UBound(SourceMat, 2))
  End Sub

  Public Shared Sub MatchVectDim(ByRef SourceVect(,) As Double, ByRef TargetVect() As Double)
    ReDim TargetVect(UBound(SourceVect))
  End Sub

  Public Shared Sub DuplicateMat(ByRef SourceMat(,) As Double, ByRef TargetMat(,) As Double)
    Dim nRows As Integer
    Dim nCols As Integer
    Dim I As Integer, J As Integer

    On Error GoTo HandleErr
    nRows = UBound(SourceMat, 1) + 1
    nCols = UBound(SourceMat, 2) + 1
    ReDim TargetMat(UBound(SourceMat, 1), UBound(SourceMat, 2))
    For I = 0 To nRows - 1
      For J = 0 To nCols - 1
        TargetMat(I, J) = SourceMat(I, J)
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "DuplicateMat")
    Resume ExitProc
  End Sub

  Public Shared Sub DuplicateVector(ByRef SourceVect() As Double, ByRef TargetVect() As Double)
    Dim nRows As Integer
    Dim I As Integer

    On Error GoTo HandleErr
    nRows = UBound(SourceVect)
    ReDim TargetVect(nRows)
    For I = 0 To nRows
      TargetVect(I) = SourceVect(I)
    Next I

ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "DuplicateVect")
    Resume ExitProc
  End Sub

  Public Shared Sub MatCopy(ByRef SourceMat(,) As Double, ByRef TargetMat(,) As Double)
    ' Copy BUT DOES NOT RESIZE target matrix
    Dim nRows As Integer
    Dim nCols As Integer
    Dim I As Integer, J As Integer

    On Error GoTo HandleErr
    nRows = UBound(SourceMat, 1) + 1
    nCols = UBound(SourceMat, 2) + 1
    For I = 0 To nRows - 1
      For J = 0 To nCols - 1
        TargetMat(I, J) = SourceMat(I, J)
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatCopy")
    Resume ExitProc
  End Sub

  Public Shared Sub MatPartialCopy(ByRef SourceMat(,) As Double, _
                                   ByRef TargetMat(,) As Double, _
                                   ByVal nRowStart As Integer, ByVal nNumRows As Integer, _
                                   ByVal nColStart As Integer, ByVal nNumCols As Integer, _
                                   ByVal nRowStartTarg As Integer, ByVal nColStartTarg As Integer)

    Dim nRowEnd As Integer
    Dim nColEnd As Integer
    Dim iCol As Integer, iRow As Integer
    Dim I As Integer
    Dim J As Integer

    On Error GoTo HandleErr
    nRowEnd = nRowStart + nNumRows - 1
    nColEnd = nColStart + nNumCols - 1

    If (nRowStart < 0) Or (nNumRows < 1) Or (nRowEnd < 0) Or _
       (nColStart < 0) Or (nNumCols < 1) Or (nColEnd < 0) Then
      ReportError("Invalid indices", "MatPartialCopy")
      Exit Sub
    End If

    iRow = nRowStartTarg
    For I = nRowStart To nRowEnd
      iCol = nColStartTarg
      For J = nColStart To nColEnd
        TargetMat(iRow, iCol) = SourceMat(I, J)
        iCol = iCol + 1
      Next J
      iRow = iRow + 1
    Next I

ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatPartialCopy")
    Resume ExitProc
  End Sub

  Public Shared Sub CopySomeVectElemnts(ByRef SourceVect() As Double, ByRef TargetVect() As Double, _
  ByVal nFirst As Integer, ByVal nNumElems As Integer, ByVal nFirstTarg As Integer)

    Dim I As Integer
    Dim nLast As Integer

    On Error GoTo HandleErr
    nLast = nFirst + nNumElems - 1

    If (nFirst < 0) Or (nNumElems < 1) Or (nLast < 0) Then
      ReportError("Invalid indices", "CopySomeVectElemnts")
      Exit Sub
    End If

    For I = nFirst To nLast
      TargetVect(nFirstTarg) = SourceVect(I)
      nFirstTarg = nFirstTarg + 1
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "CopySomeVectElemnts")
    Resume ExitProc
  End Sub

  Public Shared Function VectorFindVal(ByVal AVect() As Double, ByVal SearchVal As Double, _
                                       ByVal MinDiff As Double, ByVal nFirst As Integer) As Integer
    Dim I As Integer

    If nFirst < 0 Then nFirst = 0

    For I = nFirst To UBound(AVect)
      If Math.Abs(AVect(I) - SearchVal) < MinDiff Then
        Return I
      End If
    Next I
    Return -1
  End Function

  Public Shared Function VectorFindValRev(ByVal AVect() As Double, ByVal SearchVal As Double, _
  ByVal MinDiff As Double, ByVal nFirst As Integer) As Integer
    Dim I As Integer

    If nFirst > UBound(AVect) Then nFirst = UBound(AVect)

    For I = nFirst To 0 Step -1
      If Math.Abs(AVect(I) - SearchVal) < MinDiff Then
        Return I
      End If
    Next I
    Return -1
  End Function

  Public Shared Function MatFindValByRows(ByVal A(,) As Double, ByVal SearchVal As Double, _
         ByVal StartRow As Integer, ByVal StartCol As Integer, _
         ByRef RowFound As Integer, ByRef ColFound As Integer, ByVal MinDiff As Double) As Boolean
    ' search row by row
    Dim Row As Integer
    Dim Col As Integer

    If StartRow < 0 Then StartRow = 0
    If StartCol < 0 Then StartCol = 0
    ' search StartRow first
    For Col = StartCol To UBound(A, 2)
      If Math.Abs(A(StartRow, Col) - SearchVal) < MinDiff Then
        RowFound = StartRow
        ColFound = Col
        Return True
      End If
    Next Col
    ' search the entire subsequent rows
    For Col = 0 To UBound(A, 2)
      For Row = StartRow + 1 To UBound(A, 1)
        If Math.Abs(A(Row, Col) - SearchVal) < MinDiff Then
          RowFound = Row
          ColFound = Col
          Return True
        End If
      Next Row
    Next Col
    RowFound = -1
    ColFound = -1
    Return False
  End Function

  Public Shared Function MatFindValByRowsRev(ByVal A(,) As Double, ByVal SearchVal As Double, _
  ByVal StartRow As Integer, ByVal StartCol As Integer, _
         ByRef RowFound As Integer, ByRef ColFound As Integer, ByVal MinDiff As Double) As Boolean
    ' reverse search row by row. Starts with higher-index rows and work to
    ' low-index rows
    Dim Row As Integer
    Dim Col As Integer

    If StartRow > UBound(A, 1) Then StartRow = UBound(A, 1)
    If StartCol > UBound(A, 2) Then StartCol = UBound(A, 2)
    ' search StartRow first
    For Col = StartCol To 0 Step -1
      If Math.Abs(A(StartRow, Col) - SearchVal) < MinDiff Then
        RowFound = StartRow
        ColFound = Col
        Return True
      End If
    Next Col
    ' search the entire previous rows
    For Col = UBound(A, 2) To 0 Step -1
      For Row = StartRow - 1 To 0 Step -1
        If Math.Abs(A(Row, Col) - SearchVal) < MinDiff Then
          RowFound = Row
          ColFound = Col
          Return True
        End If
      Next Row
    Next Col
    RowFound = -1
    ColFound = -1
    Return False
  End Function


  Public Shared Function MatFindValByCols(ByVal A(,) As Double, ByVal SearchVal As Double, _
  ByVal StartRow As Integer, ByVal StartCol As Integer, _
         ByRef RowFound As Integer, ByRef ColFound As Integer, ByVal MinDiff As Double) As Boolean
    ' search column by column
    Dim Row As Integer
    Dim Col As Integer

    If StartRow < 0 Then StartRow = 0
    If StartCol < 0 Then StartCol = 0
    ' search StartRow first
    For Row = StartRow To UBound(A, 1)
      If Math.Abs(A(Row, StartCol) - SearchVal) < MinDiff Then
        RowFound = Row
        ColFound = StartCol
        Return True
      End If
    Next Row
    ' search the entire subsequent columns
    For Row = 0 To UBound(A, 1)
      For Col = StartCol + 1 To UBound(A, 2)
        If Math.Abs(A(Row, Col) - SearchVal) < MinDiff Then
          RowFound = Row
          ColFound = Col
          Return True
        End If
      Next Col
    Next Row
    RowFound = -1
    ColFound = -1
    Return False
  End Function

  Public Shared Function MatFindValByColsRev(ByVal A(,) As Double, ByVal SearchVal As Double, _
  ByVal StartRow As Integer, ByVal StartCol As Integer, _
         ByRef RowFound As Integer, ByRef ColFound As Integer, ByVal MinDiff As Double) As Boolean
    ' backward search column by column
    Dim Row As Integer
    Dim Col As Integer

    If StartRow < 0 Then StartRow = 0
    If StartCol < 0 Then StartCol = 0
    ' search StartRow first
    For Row = StartRow To 0 Step -1
      If Math.Abs(A(Row, StartCol) - SearchVal) < MinDiff Then
        RowFound = Row
        ColFound = StartCol
        Return True
      End If
    Next Row
    ' search the entire subsequent columns
    For Row = UBound(A, 1) To 0 Step -1
      For Col = StartCol - 1 To 0 Step -1
        If Math.Abs(A(Row, Col) - SearchVal) < MinDiff Then
          RowFound = Row
          ColFound = Col
          Return True
        End If
      Next Col
    Next Row
    RowFound = -1
    ColFound = -1
    Return False
  End Function


  Public Shared Sub MatSwapRows(ByRef A(,) As Double, ByVal Row1 As Integer, ByVal Row2 As Integer)
    Dim Col As Integer
    Dim X As Double

    If (Row1 < 0) Or (Row1 > UBound(A, 1)) Or _
       (Row2 < 0) Or (Row2 > UBound(A, 1)) Then
      ReportError("Invalid row values", "MatSwapRows")
      Exit Sub
    End If

    If Row1 = Row2 Then Exit Sub

    On Error GoTo HandleErr

    For Col = 0 To UBound(A, 2)
      X = A(Row1, Col)
      A(Row1, Col) = A(Row2, Col)
      A(Row2, Col) = X
    Next Col

ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatSwapRows")
    Resume ExitProc
  End Sub

  Public Shared Sub MatSwapCols(ByRef A(,) As Double, ByVal Col1 As Integer, ByVal Col2 As Integer)
    Dim Row As Integer
    Dim X As Double

    If (Col1 < 0) Or (Col1 > UBound(A, 2)) Or _
       (Col2 < 0) Or (Col2 > UBound(A, 2)) Then
      ReportError("Invalid row values", "MatSwapCols")
      Exit Sub
    End If

    If Col1 = Col2 Then Exit Sub

    On Error GoTo HandleErr

    For Row = 0 To UBound(A, 1)
      X = A(Row, Col1)
      A(Row, Col1) = A(Row, Col2)
      A(Row, Col2) = X
    Next Row

ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatSwapCols")
    Resume ExitProc
  End Sub

  Public Shared Function MatRowSum(ByRef A(,) As Double, ByVal nRow As Integer) As Double
    Dim fSum As Double
    Dim nCol As Integer

    fSum = mfBadDataVal
    On Error GoTo HandleErr
    If (nRow >= 0) And (nRow <= UBound(A, 1)) Then
      fSum = 0.0#
      For nCol = 0 To UBound(A, 2)
        fSum = fSum + A(nRow, nCol)
      Next nCol
    End If

ExitProc:
    MatRowSum = fSum
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatRowSum")
    Resume ExitProc
  End Function

  Public Shared Function MatColSum(ByRef A(,) As Double, ByVal nCol As Integer) As Double
    Dim fSum As Double
    Dim nRow As Integer

    fSum = mfBadDataVal
    On Error GoTo HandleErr
    If (nCol >= 0) And (nCol <= UBound(A, 2)) Then
      fSum = 0.0#
      For nRow = 0 To UBound(A, 1)
        fSum = fSum + A(nRow, nCol)
      Next nRow
    End If

ExitProc:
    MatColSum = fSum
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatColSum")
    Resume ExitProc
  End Function

  Public Shared Sub MatColStats(ByRef A(,) As Double, ByVal nCol As Integer, _
                      ByRef N As Integer, ByRef Mean As Double, _
                      ByRef Sdev As Double)
    Dim X As Double
    Dim nRow As Integer
    Dim SumX As Double, SumX2 As Double

    Mean = mfBadDataVal
    Sdev = mfBadDataVal
    N = -1
    On Error GoTo HandleErr
    If (nCol >= 0) And (nCol <= UBound(A, 2)) Then
      N = UBound(A, 1) + 1
      SumX = 0
      SumX2 = 0
      For nRow = 0 To UBound(A, 1)
        X = A(nRow, nCol)
        SumX = SumX + X
        SumX2 = SumX2 + X * X
      Next nRow
    End If

    Mean = SumX / N
    Sdev = Math.Sqrt((SumX2 - SumX ^ 2 / N) / (N - 1))

ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatColStats")
    Resume ExitProc
  End Sub

  Public Shared Function MatRowSumSquared(ByVal A(,) As Double, ByVal nRow As Integer) As Double
    Dim fSum As Double
    Dim nCol As Integer

    fSum = mfBadDataVal
    On Error GoTo HandleErr
    If nRow >= 0 And nRow <= UBound(A, 1) Then
      fSum = 0.0#
      For nCol = 0 To UBound(A, 2)
        fSum = fSum + A(nRow, nCol) ^ 2
      Next nCol
    End If
ExitProc:
    MatRowSumSquared = fSum
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatRowSumSquared")
    Resume ExitProc
  End Function

  Public Shared Function MatColSumSquared(ByVal A(,) As Double, ByVal nCol As Integer) As Double
    Dim fSum As Double
    Dim nRow As Integer

    fSum = mfBadDataVal
    On Error GoTo HandleErr
    If nCol >= 0 And nCol <= UBound(A, 2) Then
      fSum = 0.0#
      For nRow = 0 To UBound(A, 1)
        fSum = fSum + A(nRow, nCol) ^ 2
      Next nRow
    End If

ExitProc:
    MatColSumSquared = fSum
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatColSumSquared")
    Resume ExitProc
  End Function

  Public Shared Function MatSumDiffTwoCols(ByVal A(,) As Double, ByVal nCol1 As Integer, ByVal nCol2 As Integer) As Double
    Dim fSum As Double
    Dim nRow As Integer

    fSum = mfBadDataVal
    On Error GoTo HandleErr
    If (nCol1 >= 0) And (nCol1 <= UBound(A, 2)) And _
       (nCol2 >= 0) And (nCol2 <= UBound(A, 2)) Then
      fSum = 0.0#
      For nRow = 0 To UBound(A, 1)
        fSum = fSum + (A(nRow, nCol2) - A(nRow, nCol1))
      Next nRow
    End If

ExitProc:
    MatSumDiffTwoCols = fSum
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatSumDiffTwoCols")
    Resume ExitProc
  End Function

  Public Shared Function MatSumDiffTwoColsSquared(ByVal A(,) As Double, ByVal nCol1 As Integer, ByVal nCol2 As Integer) As Double
    Dim fSum As Double
    Dim nRow As Integer

    fSum = mfBadDataVal
    On Error GoTo HandleErr
    If (nCol1 >= 0) And (nCol1 <= UBound(A, 2)) And _
       (nCol2 >= 0) And (nCol2 <= UBound(A, 2)) Then
      fSum = 0.0#
      For nRow = 0 To UBound(A, 1)
        fSum = fSum + (A(nRow, nCol2) - A(nRow, nCol1)) ^ 2
      Next nRow
    End If

ExitProc:
    MatSumDiffTwoColsSquared = fSum
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatSumDiffTwoColsSquared")
    Resume ExitProc
  End Function


  Public Shared Function MatRowMinMax(ByRef A(,) As Double, ByVal nRow As Integer, ByRef fMax As Double) As Double
    Dim nCol As Integer
    Dim fMin As Double

    fMax = mfBadDataVal
    fMin = mfBadDataVal
    On Error GoTo HandleErr

    If (nRow >= 0) And (nRow <= UBound(A, 1)) Then
      fMin = A(nRow, 0)
      fMax = fMin
      For nCol = 1 To UBound(A, 2)

        If fMin > A(nRow, nCol) Then
          fMin = A(nRow, nCol)
        End If

        If fMax < A(nRow, nCol) Then
          fMax = A(nRow, nCol)
        End If

      Next nCol
    End If

ExitProc:
    MatRowMinMax = fMin
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatRowMinMax")
    Resume ExitProc
  End Function

  Public Shared Function MatColMinMax(ByRef A(,) As Double, ByVal nCol As Integer, ByRef fMax As Double) As Double
    Dim nRow As Integer
    Dim fMin As Double

    fMin = mfBadDataVal
    fMax = mfBadDataVal
    On Error GoTo HandleErr

    If (nCol >= 0) And (nCol <= UBound(A, 2)) Then
      fMin = A(0, nCol)
      fMax = fMin
      For nRow = 1 To UBound(A, 1)

        If fMin > A(nRow, nCol) Then
          fMin = A(nRow, nCol)
        End If

        If fMax < A(nRow, nCol) Then
          fMax = A(nRow, nCol)
        End If

      Next nRow
    End If

ExitProc:
    MatColMinMax = fMin
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatColMinMax")
    Resume ExitProc
  End Function


  Public Shared Function MatDiaglElemProduct(ByRef A(,) As Double) As Double
    Dim fProduct As Double
    Dim nRow As Integer
    Dim nCol As Integer
    Dim N As Integer

    fProduct = 1
    On Error GoTo HandleErr
    If UBound(A, 2) = UBound(A, 1) Then
      For nRow = 0 To UBound(A, 1)
        fProduct = fProduct * A(nRow, nRow)
      Next nRow
    Else
      N = IIf(UBound(A, 2) > UBound(A, 1), UBound(A, 1) + 1, UBound(A, 2) + 1)
      For nRow = 0 To N - 1
        For nCol = 0 To N - 1
          If nRow = nCol Then fProduct = fProduct * A(nRow, nCol)
        Next nCol
      Next nRow
    End If
ExitProc:
    MatDiaglElemProduct = fProduct
    Exit Function

HandleErr:
    ReportError(Err.Description, "MatDiaglElemProduct")
    Resume ExitProc
  End Function

  Public Shared Sub VectToColumnMat(ByRef AVect() As Double, ByRef A(,) As Double)
    Dim I As Integer

    ReDim A(UBound(AVect), 0)
    For I = 0 To UBound(AVect)
      A(I, 0) = AVect(I)
    Next I
  End Sub

  Public Shared Sub VectToRowMat(ByRef AVect() As Double, ByRef A(,) As Double)
    Dim I As Integer

    ReDim A(0, UBound(AVect))
    For I = 0 To UBound(AVect)
      A(0, I) = AVect(I)
    Next I
  End Sub

  Public Shared Function VectNorm(ByRef AVect() As Double, Optional ByVal P As Integer = 2) As Double
    Dim I As Integer
    Dim fNorm As Double

    fNorm = 0
    For I = 0 To UBound(AVect)
      fNorm = fNorm + AVect(I) ^ P
    Next I
    fNorm = fNorm ^ (1 / P)
    VectNorm = fNorm

  End Function


  Public Shared Function VectSum(ByRef AVect() As Double) As Double
    Dim I As Integer
    Dim Sum As Double

    Sum = 0
    For I = 0 To UBound(AVect)
      Sum = Sum + AVect(I)
    Next I
    VectSum = Sum
  End Function

  Public Shared Function VectSumAbsVal(ByRef AVect() As Double) As Double
    Dim I As Integer
    Dim Sum As Double

    Sum = 0
    For I = 0 To UBound(AVect)
      Sum = Sum + Math.Abs(AVect(I))
    Next I
    VectSumAbsVal = Sum
  End Function


  Public Shared Function VectProd(ByRef AVect() As Double) As Double
    Dim I As Integer
    Dim Prod As Double

    Prod = 1
    For I = 0 To UBound(AVect)
      Prod = Prod * AVect(I)
    Next I
    Return Prod
  End Function

  Public Shared Function VectProdOfSum(ByRef AVect() As Double, ByVal X As Double) As Double
    Dim I As Integer
    Dim Prod As Double

    Prod = 1
    For I = 0 To UBound(AVect)
      Prod = Prod * (X + AVect(I))
    Next I
    Return Prod
  End Function

  Public Shared Function VectProdOfXminusVect(ByRef AVect() As Double, ByVal X As Double) As Double
    Dim I As Integer
    Dim Prod As Double

    Prod = 1
    For I = 0 To UBound(AVect)
      Prod = Prod * (X - AVect(I))
    Next I
    Return Prod
  End Function

  Public Shared Function VectProdOfVectMinusX(ByRef AVect() As Double, ByVal X As Double) As Double
    Dim I As Integer
    Dim Prod As Double

    Prod = 1
    For I = 0 To UBound(AVect)
      Prod = Prod * (AVect(I) - X)
    Next I
    VectProdOfVectMinusX = Prod
  End Function

  Public Shared Function MatNorm(ByRef A(,) As Double, Optional ByVal P As Integer = 2) As Double
    Dim I As Integer
    Dim J As Integer
    Dim fNorm As Double

    fNorm = 0
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        fNorm = fNorm + A(I, J) ^ P
      Next J
    Next I
    fNorm = fNorm ^ (1 / P)
    MatNorm = fNorm

  End Function

  Public Shared Sub NormalizeVect(ByRef AVect() As Double)
    Dim I As Integer
    Dim fNorm As Double

    fNorm = 0
    For I = 0 To UBound(AVect)
      fNorm = fNorm + AVect(I) ^ 2
    Next I
    fNorm = Math.Sqrt(fNorm)
    For I = 0 To UBound(AVect)
      AVect(I) = AVect(I) / fNorm
    Next I
  End Sub

  Public Shared Sub MatAddMat(ByRef A(,) As Double, ByRef B(,) As Double, ByRef C(,) As Double)
    Dim I As Integer, J As Integer

    ReDim C(UBound(A, 1), UBound(A, 2))
    On Error GoTo HandleErr
    For I = LBound(A, 1) To UBound(A, 1)
      For J = LBound(A, 2) To UBound(A, 2)
        C(I, J) = A(I, J) + B(I, J)
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatAddMat")
    Resume ExitProc
  End Sub      'Add

  Public Shared Sub MatAddMatInPlace(ByRef A(,) As Double, ByRef B(,) As Double)
    Dim I As Integer, J As Integer

    On Error GoTo HandleErr
    For I = LBound(A, 1) To UBound(A, 1)
      For J = LBound(A, 2) To UBound(A, 2)
        A(I, J) = A(I, J) + B(I, J)
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatAddMatInPlace")
    Resume ExitProc
  End Sub      'Add


  Public Shared Sub MatSubVal(ByVal X As Double, ByRef A(,) As Double, ByRef C(,) As Double)
    Dim I As Integer, J As Integer

    ReDim C(UBound(A, 1), UBound(A, 2))
    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        C(I, J) = A(I, J) - X
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReDim C(1, 1)
    C(1, 1) = mfBadDataVal
    ReportError(Err.Description, "MatSubVal")
    Resume ExitProc
  End Sub      'Sub

  Public Shared Sub MatSubValInPlace(ByVal X As Double, ByRef A(,) As Double)
    Dim I As Integer, J As Integer

    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = A(I, J) - X
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatSubValInPlace")
    Resume ExitProc
  End Sub      'Sub


  Public Shared Sub MatSubMat(ByRef A(,) As Double, ByRef B(,) As Double, ByRef C(,) As Double)
    Dim I As Integer, J As Integer

    ReDim C(UBound(A, 1), UBound(A, 2))
    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        C(I, J) = A(I, J) - B(I, J)
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReDim C(1, 1)
    C(1, 1) = mfBadDataVal
    ReportError(Err.Description, "MatSubMat")
    Resume ExitProc
  End Sub

  Public Shared Sub MatSubMatInPlace(ByRef A(,) As Double, ByRef B(,) As Double)
    Dim I As Integer, J As Integer

    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = A(I, J) - B(I, J)
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatSubMatInPlace")
    Resume ExitProc
  End Sub


  Public Shared Sub MatAddVal(ByRef A(,) As Double, ByVal X As Double, ByRef C(,) As Double)
    Dim I As Integer, J As Integer

    ReDim C(UBound(A, 1), UBound(A, 2))
    On Error GoTo HandleErr
    For I = LBound(A, 1) To UBound(A, 1)
      For J = LBound(A, 2) To UBound(A, 2)
        C(I, J) = A(I, J) + X
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReDim C(1, 1)
    C(1, 1) = mfBadDataVal
    ReportError(Err.Description, "MatAddVal")
    Resume ExitProc
  End Sub

  Public Shared Sub MatAddValInPlace(ByRef A(,) As Double, ByVal X As Double)
    Dim I As Integer, J As Integer

    On Error GoTo HandleErr
    For I = LBound(A, 1) To UBound(A, 1)
      For J = LBound(A, 2) To UBound(A, 2)
        A(I, J) = A(I, J) + X
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatAddValInPlace")
    Resume ExitProc
  End Sub

  Public Shared Sub VectAddVect(ByRef AVect() As Double, ByRef BVect() As Double, _
                         ByRef CVect() As Double)
    Dim I As Integer

    ReDim CVect(UBound(AVect))
    For I = 0 To UBound(AVect)
      CVect(I) = AVect(I) + BVect(I)
    Next I

  End Sub

  Public Shared Sub VectAddScaledVect(ByRef AVect() As Double, ByVal X As Double, _
                         ByRef BVect() As Double, ByRef CVect() As Double)
    Dim I As Integer

    ReDim CVect(UBound(AVect))
    For I = 0 To UBound(AVect)
      CVect(I) = AVect(I) + X * BVect(I)
    Next I

  End Sub


  Public Shared Function VectTransMultVect(ByRef AVect() As Double, ByRef BVect() As Double) As Double
    Dim Res As Double
    Dim I As Integer

    Res = 0
    For I = 0 To UBound(AVect)
      Res = Res + AVect(I) * BVect(I)
    Next I
    VectTransMultVect = Res
  End Function

  Public Shared Sub VectMultVal(ByRef AVect() As Double, ByVal X As Double, ByRef BVect() As Double)
    Dim I As Integer

    ReDim BVect(UBound(AVect))
    For I = 0 To UBound(AVect)
      BVect(I) = X * AVect(I)
    Next I

  End Sub

  Public Shared Sub VectMultValInPlace(ByRef AVect() As Double, ByVal X As Double)
    Dim I As Integer

    For I = 0 To UBound(AVect)
      AVect(I) = AVect(I) * X
    Next I

  End Sub

  Public Shared Sub MatMultVal(ByRef A(,) As Double, ByRef B As Double, ByRef C(,) As Double)
    Dim I As Integer, J As Integer

    ReDim C(UBound(A, 1), UBound(A, 2))
    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        C(I, J) = B * A(I, J)
      Next J
    Next I

ExitProc:
    Exit Sub

HandleErr:
    ReDim C(1, 1)
    C(1, 1) = mfBadDataVal
    ReportError(Err.Description, "MatMultVal")
    Resume ExitProc
  End Sub

  Public Shared Sub MatMultValInPlace(ByRef A(,) As Double, ByRef B As Double)
    Dim I As Integer, J As Integer

    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = B * A(I, J)
      Next J
    Next I

ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatMultValInPlace")
    Resume ExitProc
  End Sub


  Public Shared Sub MatMultVect(ByRef A(,) As Double, ByRef BVect() As Double, ByRef CVect() As Double)
    Dim X As Double
    Dim I As Integer, J As Integer

    ReDim CVect(UBound(BVect))

    On Error GoTo HandleErr
    For J = 0 To UBound(BVect)
      CVect(J) = mfBadDataVal
    Next J
    If UBound(A, 2) <> UBound(BVect) Then
      ReportError("Incompatible matrix dimensions", "MatMultVect")
      Exit Sub
    End If

    For I = 0 To UBound(A, 1)
      X = 0
      For J = 0 To UBound(BVect)
        X = X + BVect(J) * A(I, J)
      Next J
      CVect(I) = X
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReDim CVect(1)
    CVect(1) = mfBadDataVal
    ReportError(Err.Description, "MatMultVect")
    Resume ExitProc
  End Sub

  Public Shared Sub MatMultVectInPlace(ByRef A(,) As Double, ByRef BVect() As Double)
    Dim X As Double
    Dim I As Integer, J As Integer
    Dim C() As Double

    ReDim C(UBound(BVect))
    On Error GoTo HandleErr
    For J = 0 To UBound(BVect)
      C(J) = mfBadDataVal
    Next J
    If UBound(A, 2) <> UBound(BVect) Then
      ReportError("Incompatible matrix dimensions", "MatMultVect")
      Exit Sub
    End If

    For I = 0 To UBound(A, 1)
      X = 0
      For J = 0 To UBound(BVect)
        X = X + BVect(J) * A(I, J)
      Next J
      C(I) = X
    Next I
ExitProc:
    For J = 0 To UBound(BVect)
      BVect(J) = C(J)
    Next J
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatMultVectInPlace")
    Resume ExitProc
  End Sub

  Public Shared Sub MatMultMat(ByRef A(,) As Double, _
          ByRef B(,) As Double, ByRef C(,) As Double)
    Dim fSum As Double
    Dim I As Integer, J As Integer, K As Integer

    ReDim C(UBound(A, 1), UBound(B, 2))
    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(B, 2)
        C(I, J) = mfBadDataVal
      Next J
    Next I
    If UBound(A, 2) <> UBound(B, 1) Then
      ReportError("Incompatible matrix dimensions", "MatMultMat)")
      Exit Sub
    End If
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(B, 2)
        fSum = 0
        For K = 0 To UBound(A, 2)
          fSum = fSum + A(I, K) * B(K, J)
        Next K
        C(I, J) = fSum
      Next J
    Next I

ExitProc:
    Exit Sub

HandleErr:
    ReDim C(1, 1)
    C(1, 1) = mfBadDataVal
    ReportError(Err.Description, "MatMultMat")
    Resume ExitProc
  End Sub

  Public Shared Sub MatMultMatInPlace(ByRef A(,) As Double, ByRef B(,) As Double)
    Dim fSum As Double
    Dim I As Integer, J As Integer, K As Integer
    Dim C(,) As Double

    ReDim C(UBound(A, 1), UBound(B, 2))
    On Error GoTo HandleErr
    If UBound(A, 2) <> UBound(B, 1) Then
      ReportError("Incompatible matrix dimensions", "MatMultMat)")
      Exit Sub
    End If
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(B, 2)
        fSum = 0
        For K = 0 To UBound(A, 2)
          fSum = fSum + A(I, K) * B(K, J)
        Next K
        C(I, J) = fSum
      Next J
    Next I

ExitProc:
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(B, 2)
        A(I, J) = C(I, J)
      Next J
    Next I
    Exit Sub

HandleErr:
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(B, 2)
        C(I, J) = mfBadDataVal
      Next J
    Next I
    ReportError(Err.Description, "MatMultMatInPlace")
    Resume ExitProc
  End Sub


  Public Shared Sub MatDivVal(ByRef A(,) As Double, ByVal X As Double, ByRef C(,) As Double)
    Dim I As Integer, J As Integer

    ReDim C(UBound(A, 1), UBound(A, 2))
    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        C(I, J) = A(I, J) / X
      Next J
    Next I

ExitProc:
    Exit Sub

HandleErr:
    ReDim C(1, 1)
    C(1, 1) = mfBadDataVal
    ReportError(Err.Description, "MatDivVal")
    Resume ExitProc
  End Sub

  Public Shared Sub MatDivValInPlace(ByRef A(,) As Double, ByVal X As Double)
    Dim I As Integer, J As Integer

    On Error GoTo HandleErr
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = A(I, J) / X
      Next J
    Next I

ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "MatDivValInPlace")
    Resume ExitProc
  End Sub


  Public Shared Sub MatTrans(ByRef A(,) As Double, ByRef C(,) As Double)
    Dim I As Integer, J As Integer

    ' redimension matrix C to have reverse dimensions of matrix A
    ReDim C(UBound(A, 2), UBound(A, 1))
    On Error GoTo HandleErr
    ' copy and transpose data into temp matrix x
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        C(J, I) = A(I, J)
      Next J
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReDim C(1, 1)
    C(1, 1) = mfBadDataVal
    ReportError(Err.Description, "MatTrans")
    Resume ExitProc
  End Sub      'MatTrans

  Public Shared Sub MatIdentity(ByRef A(,) As Double)
    Dim I As Integer, J As Integer

    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = IIf(I = J, 1, 0)
      Next J
    Next I
  End Sub

  Public Shared Sub MatZero(ByRef A(,) As Double)
    Dim I As Integer, J As Integer

    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = 0
      Next J
    Next I
  End Sub

  Public Shared Sub MatRandom(ByRef A(,) As Double, ByVal Xlow As Double, ByVal XHi As Double)
    Dim I As Integer, J As Integer

    Randomize(Timer)
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = Xlow + (XHi - Xlow) * Rnd(1)
      Next J
    Next I
  End Sub

  Public Shared Sub MatRandomNormal(ByRef A(,) As Double, ByVal Mean As Double, ByVal Sdev As Double)
    Dim I As Integer, J As Integer

    Randomize(Timer)
    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = RandNorm(Mean, Sdev)
      Next J
    Next I
  End Sub

  Private Shared Function RandNorm(ByVal Mean As Double, ByVal Sdev As Double) As Double
    ' RandNorm returns a randomly distributed drawing from a
    ' standard normal distribution i.e. one with:
    ' Average = 0 and Standard Deviation = 1.0
    Dim fac As Double, rsq As Double, v1 As Double, v2 As Double
    Static flag As Boolean, gset As Double

    ' Each pass through the calculation of the routine produces
    '  two normally-distributed deviates, so we only need to do
    '  the calculations every other call. So we set the flag
    '  variable (to true) if gset contains a spare RandNorm value.
    If flag Then
      RandNorm = Mean + Sdev * gset
      ' Force calculation next time.
      flag = False
    Else
      ' Don't have anything saved so need to find a pair of values
      ' First generate a co-ordinate pair within the unit circle:
      Do
        v1 = 2 * Rnd(1) - 1.0#
        v2 = 2 * Rnd(1) - 1.0#
        rsq = v1 * v1 + v2 * v2
      Loop Until rsq <= 1.0#

      ' Do the Math:
      fac = Math.Sqrt(-2.0# * Math.Log(rsq) / rsq)

      ' Return one of the values and save the other (gset) for next time:
      Return Mean + Sdev * (v2 * fac)
      gset = v1 * fac
      flag = True
    End If

  End Function

  Public Shared Sub MatFill(ByRef A(,) As Double, ByVal X As Double)
    Dim I As Integer, J As Integer

    For I = 0 To UBound(A, 1)
      For J = 0 To UBound(A, 2)
        A(I, J) = X
      Next J
    Next I
  End Sub

  '---------------------------------------------------------------------------
  ' ------------------------- Section for Sepcial Mat/Vect operations ---------
  '---------------------------------------------------------------------------
  Public Shared Sub OpMult_Vect_VectT(ByRef AVect() As Double, ByRef BVect() As Double, _
                               ByRef CMat(,) As Double)
    Dim I As Integer, J As Integer, N As Integer

    N = UBound(AVect)
    If N = UBound(BVect) Then
      ReDim CMat(N, N)
      For I = 0 To N
        For J = 0 To N
          CMat(I, J) = AVect(I) * BVect(J)
        Next J
      Next I
    Else
      ReDim CMat(0, 0)
      CMat(0, 0) = -1.1E+30
    End If
  End Sub

  Public Shared Function OpMult_VectT_Vect(ByRef AVect() As Double, ByRef BVect() As Double) As Double
    Dim Res As Double
    Dim I As Integer, N As Integer

    N = UBound(AVect)
    If N = UBound(BVect) Then
      Res = 0
      For I = 0 To N
        Res = Res + AVect(I) * BVect(I)
      Next I
    Else
      Res = -1.1E+30
    End If
    Return Res
  End Function

  Public Shared Function OpMult_VectT_Mat_SameVect(ByRef BVect() As Double, ByRef AMat(,) As Double) As Double
    Dim Res As Double
    Dim CVect() As Double
    Dim I, N As Integer

    N = UBound(BVect)
    If N = UBound(AMat, 1) And N = UBound(AMat, 2) Then
      ReDim CVect(N)
      MatMultVect(AMat, BVect, CVect)
      Res = 0
      For I = 0 To N
        Res = Res + CVect(I) * BVect(I)
      Next I
    Else
      Res = -1.1E+30
    End If
    Return Res
  End Function

  Public Shared Sub OpMult_Vect_VectT_Mat(ByRef AVect() As Double, ByRef BVect() As Double, _
                                   ByRef AMat(,) As Double, ByRef CMat(,) As Double)
    Dim N As Integer
    Dim BMat(,) As Double

    ReDim BMat(1, 1)
    N = UBound(AVect)
    If N = UBound(BVect) And N = UBound(AMat, 1) And N = UBound(AMat, 2) Then
      OpMult_Vect_VectT(AVect, BVect, BMat)
      MatMultMat(AMat, BMat, CMat)
    Else
      ReDim CMat(0, 0)
      CMat(0, 0) = -1.1E+30
    End If
  End Sub

  Public Shared Sub OpMult_Mat_Vect_VectT(ByRef AVect() As Double, ByRef BVect() As Double, _
                                   ByRef AMat(,) As Double, ByRef CMat(,) As Double)
    OpMult_Vect_VectT_Mat(AVect, BVect, AMat, CMat)
  End Sub

  Private Shared Function Pythag(ByVal A As Double, ByVal B As Double) As Double
    Dim AbsA As Double, AbsB As Double

    AbsA = Math.Abs(A)
    AbsB = Math.Abs(B)
    If AbsA > AbsB Then
      Pythag = AbsA * Math.Sqrt((1.0# + (AbsB / AbsA) ^ 2))
    Else
      Pythag = IIf(AbsB = 0.0#, 0.0#, AbsB * Math.Sqrt((1.0# + (AbsA / AbsB) ^ 2)))
    End If
  End Function

  Public Shared Sub LuDecomp(ByRef A(,) As Double, ByVal nNumRows As Integer, ByRef pivotIndex() As Integer, ByRef Factor As Double)
    Const Epsilon As Double = 1.0E-20
    Dim VV() As Double
    Dim iMax As Integer
    Dim I As Integer, J As Integer, K As Integer
    Dim big As Double, dum As Double, fSum As Double, temp As Double

    iMax = 0
    On Error GoTo HandleErr
    ReDim VV(nNumRows)
    Factor = 1.0#
    For I = 0 To nNumRows - 1
      big = 0.0#
      For J = 0 To nNumRows - 1
        temp = Math.Abs(A(I, J))
        If temp > big Then 'ToDo: Unsupported feature: assignment within expression.
          big = temp
        End If
      Next J
      If big = 0.0# Then
        ReportError("Singular matrix", "LuDecomp")
        Exit Sub
      End If
      VV(I) = 1.0# / big
    Next I
    For J = 0 To nNumRows - 1
      For I = 0 To J - 1
        fSum = A(I, J)
        For K = 0 To I - 1
          fSum = fSum - A(I, K) * A(K, J)
        Next K
        A(I, J) = fSum
      Next I
      big = 0.0#
      For I = J To nNumRows - 1
        fSum = A(I, J)
        For K = 0 To J - 1
          fSum = fSum - A(I, K) * A(K, J)
        Next K
        A(I, J) = fSum
        dum = VV(I) * Math.Abs(fSum)
        If dum >= big Then
          big = dum
          iMax = I
        End If
      Next I
      If J <> iMax Then
        For K = 0 To nNumRows - 1
          dum = A(iMax, K)
          A(iMax, K) = A(J, K)
          A(J, K) = dum
        Next K
        Factor = -Factor
        VV(iMax) = VV(J)
      End If
      pivotIndex(J) = iMax
      If A(J, J) = 0.0# Then
        A(J, J) = Epsilon
      End If
      If J <> nNumRows - 1 Then
        dum = 1.0# / A(J, J)
        For I = J + 1 To nNumRows - 1
          A(I, J) = dum * A(I, J)
        Next I
      End If
    Next J
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "LuDecomp")
    Resume ExitProc
  End Sub


  Public Shared Sub LuBackSubst(ByRef A(,) As Double, ByVal nNumRows As Integer, ByRef pivotIndex() As Integer, ByRef BVect() As Double)
    Dim I1 As Integer
    Dim I As Integer, ip As Integer, J As Integer
    Dim fSum As Double

    I1 = 0
    On Error GoTo HandleErr
    For I = 0 To nNumRows - 1
      ip = pivotIndex(I)
      fSum = BVect(ip)
      BVect(ip) = BVect(I)
      If I1 <> 0 Then
        For J = I1 - 1 To I - 1
          fSum = fSum - A(I, J) * BVect(J)
        Next J
      ElseIf fSum <> 0.0# Then
        I1 = I + 1
      End If
      BVect(I) = fSum
    Next I
    For I = nNumRows - 1 To 0 Step -1
      fSum = BVect(I)
      For J = I + 1 To nNumRows - 1
        fSum = fSum - A(I, J) * BVect(J)
      Next J
      BVect(I) = fSum / A(I, I)
    Next I
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "LuBackSubst")
    Resume ExitProc
  End Sub

  Public Shared Sub MV_LUBackSubst(ByRef A(,) As Double, ByRef Index() As Integer, _
                                   ByVal nNumRows As Integer, ByRef B() As Double)

    Dim I As Integer, J As Integer, K As Integer, idx As Integer
    Dim Sum As Double

    K = -1
    For I = 0 To nNumRows - 1
      idx = Index(I)
      Sum = B(idx)
      B(idx) = B(I)
      If K > -1 Then
        For J = K To I - 1
          Sum = Sum - A(I, J) * B(J)
        Next J
      End If
      If Sum <> 0 Then
        K = I
      End If
      B(I) = Sum
    Next I

    For I = nNumRows - 1 To 0 Step -1
      Sum = B(I)
      For J = I + 1 To nNumRows - 1
        Sum = Sum - A(I, J) * B(J)
      Next J
      B(I) = Sum / A(I, I)
    Next I
  End Sub

  Public Shared Sub MV_LUDecomp(ByRef A(,) As Double, ByRef Index() As Integer, ByVal nNumRows As Integer)

    Dim I As Integer, J As Integer, K As Integer, iMax As Integer
    Dim large As Double, Sum As Double, z As Double, Z2 As Double
    Dim scaleVect(50) As Double

    ' loop to obtain the scaling element
    For I = 0 To nNumRows - 1
      large = 0
      For J = 0 To nNumRows - 1
        Z2 = Math.Abs(A(I, J))
        If Z2 > large Then large = Z2
      Next J
      ' no non-zero large value? then exit with an error code
      If (large = 0) Then
        ReportError("Calculations error", "MV_LUDecomp")
        Exit Sub
      End If
      scaleVect(I) = 1 / large
    Next I

    For J = 0 To nNumRows - 1
      For I = 0 To J - 1
        Sum = A(I, J)
        For K = 0 To I - 1
          Sum = Sum - A(I, K) * A(K, J)
        Next K
        A(I, J) = Sum
      Next I

      large = 0

      For I = J To nNumRows - 1
        Sum = A(I, J)
        For K = 0 To J - 1
          Sum = Sum - A(I, K) * A(K, J)
        Next K
        A(I, J) = Sum
        z = scaleVect(I) * Math.Abs(Sum)
        If z >= large Then
          large = z
          iMax = I
        End If
      Next I

      If J <> iMax Then
        For K = 0 To nNumRows - 1
          z = A(iMax, K)
          A(iMax, K) = A(J, K)
          A(J, K) = z
        Next K
        scaleVect(iMax) = scaleVect(J)
      End If

      Index(J) = iMax

      If A(J, J) = 0 Then A(J, J) = 0.000000000000001

      If J <> nNumRows Then
        z = 1 / A(J, J)
        For I = J + 1 To nNumRows - 1
          A(I, J) = A(I, J) * z
        Next I
      End If
    Next J
  End Sub

  Public Shared Function MV_LUDeterminant(ByRef A(,) As Double, ByVal nNumRows As Integer, ByVal rowSwapFlag As Integer) As Double

    Dim result As Double
    Dim I As Integer

    result = rowSwapFlag
    For I = 0 To nNumRows - 1
      result = result * A(I, I)
    Next I

    Return result
  End Function

  Public Shared Sub MV_LUInverse(ByRef A(,) As Double, ByRef InvA(,) As Double, ByRef Index() As Integer, ByVal nNumRows As Integer)

    Dim colVect() As Double
    Dim I As Integer, J As Integer

    ReDim colVect(nNumRows)
    For J = 0 To nNumRows - 1
      For I = 0 To nNumRows - 1
        colVect(I) = 0
      Next I
      colVect(J) = 1
      MV_LUBackSubst(A, Index, nNumRows, colVect)
      For I = 0 To nNumRows - 1
        InvA(I, J) = colVect(I)
      Next I
    Next J

  End Sub

  Public Shared Function MatDeterminant(ByRef A(,) As Double, ByVal nNumRows As Integer) As Double

    Dim Index() As Integer
    Dim I As Integer
    Dim rowSwapFlag As Integer
    Dim result As Double
    Dim nErrCount As Integer

    ReDim Index(nNumRows)
    nErrCount = GetErrorCount
    MV_LUDecomp(A, Index, nNumRows)
    If nErrCount <> GetErrorCount Then
      MatDeterminant = mfBadDataVal
      Exit Function
    End If

    result = rowSwapFlag

    For I = 0 To nNumRows - 1
      result = result * A(I, I)
    Next I
    Return result
  End Function

  Public Shared Sub MatInverse(ByRef A(,) As Double, ByVal nNumRows As Integer)

    Dim colVect() As Double
    Dim Index() As Integer
    Dim I As Integer, J As Integer
    Dim nErrCount As Integer

    ReDim colVect(nNumRows)
    ReDim Index(nNumRows)
    nErrCount = GetErrorCount
    MV_LUDecomp(A, Index, nNumRows)
    If nErrCount <> GetErrorCount Then
      ReportError("Error", "MatInverse")
      Exit Sub
    End If

    For J = 0 To nNumRows - 1
      For I = 0 To nNumRows - 1
        colVect(I) = 0
      Next I
      colVect(J) = 1
      MV_LUBackSubst(A, Index, nNumRows, colVect)
      For I = 0 To nNumRows - 1
        A(I, J) = colVect(I)
      Next I
    Next J

  End Sub


  Public Shared Sub SolveLU(ByRef A(,) As Double, ByRef BVect() As Double, ByVal nCount As Integer)
    Dim pvtIdx() As Integer
    Dim Factor As Double
    Dim nInitErrCount As Integer

    ReDim pvtIdx(nCount)
    Factor = 1
    nInitErrCount = GetErrorCount()

    LuDecomp(A, nCount, pvtIdx, Factor)
    ' no new errors?
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuBackSubst(A, nCount, pvtIdx, BVect)
  End Sub


  Public Shared Sub LuImprove(ByRef A(,) As Double, ByRef ALud(,) As Double, ByRef pivotIndex() As Integer, _
         ByRef BVect() As Double, ByRef Xvect() As Double, _
         ByVal nCount As Integer)
    Dim I, J As Integer
    Dim sdp As Double
    Dim R() As Double

    ReDim R(nCount)
    For I = 0 To nCount - 1
      sdp = -BVect(I)
      For J = 0 To nCount - 1
        sdp = sdp + A(I, J) * Xvect(J)
      Next J
      R(I) = sdp
    Next I
    LuBackSubst(ALud, nCount, pivotIndex, R)
    For I = 0 To nCount - 1
      Xvect(I) = Xvect(I) - R(I)
    Next I
  End Sub

  Public Shared Sub SolveLU_Ex(ByRef A(,) As Double, ByRef B() As Double, ByVal nCount As Integer)
    Dim ACopy(,) As Double
    Dim BCopy() As Double
    Dim pvtIdx() As Integer
    Dim Factor As Double
    Dim nInitErrCount As Integer

    Factor = 1
    nInitErrCount = GetErrorCount()
    ReDim BCopy(UBound(B))
    ReDim pvtIdx(nCount)
    ReDim ACopy(1, 1)
    ACopy(0, 0) = 0

    DuplicateMat(A, ACopy)
    DuplicateVector(B, BCopy)

    LuDecomp(A, nCount, pvtIdx, Factor)
    ' no new errors?
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuImprove(ACopy, A, pvtIdx, BCopy, B, nCount)
  End Sub 'SolveLU_Ex


  Public Shared Sub SvdDecomp(ByRef A(,) As Double, ByVal nNumRows As Integer, ByVal nNumCols As Integer, ByRef W() As Double, ByRef V(,) As Double)
    Const MAX_ITERS As Integer = 30
    Dim rv1() As Double
    Dim flag As Boolean
    Dim L As Integer
    Dim nm As Integer
    Dim IterCount As Integer, I As Integer
    Dim J As Integer, M As Integer, K As Integer
    Dim C As Double, f As Double, h As Double, s As Double
    Dim X As Double, Y As Double, z As Double
    Dim g As Double
    Dim fScale As Double
    Dim Anorm As Double

    ReDim rv1(nNumRows - 1)
    L = 0
    nm = 0
    g = 0
    fScale = 0
    Anorm = 0

    For I = 0 To nNumRows - 1
      L = I + 2
      rv1(I) = fScale * g
      g = 0
      s = 0
      fScale = 0.0#
      If I < nNumCols Then
        For K = I To nNumCols - 1
          fScale = fScale + Math.Abs(A(K, I))
        Next K
        If fScale <> 0.0# Then
          For K = I To nNumCols - 1
            A(K, I) = A(K, I) / fScale
            s = s + A(K, I) ^ 2
          Next K
          f = A(I, I)
          g = -Sign(Math.Sqrt(s), f)
          h = f * g - s
          A(I, I) = f - g
          For J = L - 1 To nNumRows - 1
            s = 0
            For K = I To nNumCols - 1
              s = s + A(K, I) * A(K, J)
            Next K
            f = s / h
            For K = I To nNumCols - 1
              A(K, J) = A(K, J) + f * A(K, I)
            Next K
          Next J
          For K = I To nNumCols - 1
            A(K, I) = fScale * A(K, I)
          Next K
        End If
      End If
      W(I) = fScale * g
      g = 0
      s = 0
      fScale = 0.0#
      If I + 1 <= nNumCols And I <> nNumRows Then
        For K = L - 1 To nNumRows - 1
          fScale = fScale + Math.Abs(A(I, K))
        Next K
        If fScale <> 0.0# Then
          For K = L - 1 To nNumRows - 1
            A(I, K) = A(I, K) / fScale
            s = s + A(I, K) * A(I, K)
          Next K
          f = A(I, L - 1)
          g = -Sign(Math.Sqrt(s), f)
          h = f * g - s
          A(I, L - 1) = f - g
          For K = L - 1 To nNumRows - 1
            rv1(K) = A(I, K) / h
          Next K
          For J = L - 1 To nNumCols - 1
            s = 0
            For K = L - 1 To nNumRows - 1
              s = s + A(J, K) * A(I, K)
            Next K
            For K = L - 1 To nNumRows - 1
              A(J, K) = A(J, K) + s * rv1(K)
            Next K
          Next J
          For K = L - 1 To nNumRows - 1
            A(I, K) = fScale * A(I, K)
          Next K
        End If
      End If '
      Anorm = Max(Anorm, Math.Abs(W(I)) + Math.Abs(rv1(I)))
    Next I
    For I = nNumRows - 1 To 0 Step -1
      If I < nNumRows - 1 Then
        If g <> 0.0# Then
          For J = L To nNumRows - 1
            V(J, I) = A(I, J) / A(I, L) / g
          Next J
          For J = L To nNumRows - 1
            s = 0
            For K = L To nNumRows - 1
              s = s + A(I, K) * V(K, J)
            Next K
            For K = L To nNumRows - 1
              V(K, J) = V(K, J) + s * V(K, I)
            Next K
          Next J
        End If
        For J = L To nNumRows - 1
          V(I, J) = 0.0#
          V(J, I) = 0.0#
        Next J
      End If
      V(I, I) = 1.0#
      g = rv1(I)
      L = I
    Next I
    For I = IIf(nNumCols < nNumRows, nNumCols, nNumRows) - 1 To 0 Step -1
      L = I + 1
      g = W(I)
      For J = L To nNumRows - 1
        A(I, J) = 0.0#
      Next J
      If g <> 0.0# Then
        g = 1.0# / g
        For J = L To nNumRows - 1
          s = 0
          For K = L To nNumCols - 1
            s = s + A(K, I) * A(K, J)
          Next K
          f = s / A(I, I) * g
          For K = I To nNumCols - 1
            A(K, J) = A(K, J) + f * A(K, I)
          Next K
        Next J
        For J = I To nNumCols - 1
          A(J, I) = g * A(J, I)
        Next J
      Else
        For J = I To nNumCols - 1
          A(J, I) = 0.0#
        Next J
      End If
      A(I, I) = A(I, I) + 1
    Next I
    For K = nNumRows - 1 To 0 Step -1
      For IterCount = 0 To 29
        flag = True
        For L = K To 0 Step -1
          nm = L - 1
          If Math.Abs(rv1(L)) + Anorm = Anorm Then
            flag = False
            Exit For
          End If
          If Math.Abs(W(nm)) + Anorm = Anorm Then
            Exit For
          End If
        Next L
        If flag Then
          C = 0.0#
          s = 1.0#
          For I = L - 1 To (K + 1) - 1
            f = s * rv1(I)
            rv1(I) = C * rv1(I)
            If Math.Abs(f) + Anorm = Anorm Then
              Exit For
            End If
            g = W(I)
            h = Pythag(f, g)
            W(I) = h
            h = 1.0# / h
            C = g * h
            s = -f * h
            For J = 0 To nNumCols - 1
              Y = A(J, nm)
              z = A(J, I)
              A(J, nm) = Y * C + z * s
              A(J, I) = z * C - Y * s
            Next J
          Next I
        End If
        z = W(K)
        If L = K Then
          If z < 0.0# Then
            W(K) = -z
            For J = 0 To nNumRows - 1
              V(J, K) = -V(J, K)
            Next J
          End If
          Exit For
        End If
        If IterCount = MAX_ITERS Then
          ReportError("Failed to converge in " & IterCount & " IterCount", "SvdDecomp")
          Exit Sub
        End If
        X = W(L)
        nm = K - 1
        Y = W(nm)
        g = rv1(nm)
        h = rv1(K)
        f = ((Y - z) * (Y + z) + (g - h) * (g + h)) / (2.0# * h * Y)
        g = Pythag(f, 1.0#)
        f = ((X - z) * (X + z) + h * (Y / (f + Sign(g, f)) - h)) / X
        C = 1.0#
        s = 1.0#
        For J = L To nm
          I = J + 1
          g = rv1(I)
          Y = W(I)
          h = s * g
          g = C * g
          z = Pythag(f, h)
          rv1(J) = z
          C = f / z
          s = h / z
          f = X * C + g * s
          g = g * C - X * s
          h = Y * s
          Y = Y * C
          For M = 0 To nNumRows - 1
            X = V(M, J)
            z = V(M, I)
            V(M, J) = X * C + z * s
            V(M, I) = z * C - X * s
          Next M
          z = Pythag(f, h)
          W(J) = z
          If z <> 0 Then
            z = 1.0# / z
            C = f * z
            s = h * z
          End If
          f = C * g + s * Y
          X = C * Y - s * g
          For M = 0 To nNumCols - 1
            Y = A(M, J)
            z = A(M, I)
            A(M, J) = Y * C + z * s
            A(M, I) = z * C - Y * s
          Next M
        Next J
        rv1(L) = 0.0#
        rv1(K) = f
        W(K) = X
      Next IterCount
    Next K
  End Sub

  Public Shared Sub SvdBackSubst(ByRef U(,) As Double, ByRef W() As Double, ByRef V(,) As Double, ByRef BVect() As Double, ByRef Xvect() As Double, ByVal nNumRows As Integer, ByVal nNumCols As Integer)
    Dim M As Integer, J As Integer, I As Integer
    Dim s As Double
    Dim tmp() As Double

    ReDim tmp(nNumRows - 1)
    ReDim Xvect(nNumRows - 1)
    For J = 0 To nNumRows - 1
      s = 0.0#
      If W(J) <> 0.0# Then
        For I = 0 To nNumCols - 1
          s = s + U(I, J) * BVect(I)
        Next I
        s = s / W(J)
      End If
      tmp(J) = s
    Next J
    For J = 0 To nNumRows - 1
      s = 0.0#
      For M = 0 To nNumRows - 1
        s = s + V(J, M) * tmp(M)
      Next M
      Xvect(J) = s
    Next J
  End Sub

  Public Shared Sub SolveSvd(ByRef A(,) As Double, ByRef BVect() As Double, _
           ByRef Xvect() As Double, ByVal nCount As Integer)
    Dim V(,) As Double
    Dim W() As Double
    Dim nInitErrCount As Integer

    ReDim W(nCount - 1)
    ReDim V(nCount - 1, nCount - 1)
    nInitErrCount = GetErrorCount()

    SvdDecomp(A, nCount, nCount, W, V)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    SvdBackSubst(A, W, V, BVect, Xvect, nCount, nCount)
  End Sub 'SolveSvd

  Public Shared Sub QrDecomp(ByRef A(,) As Double, ByRef CVect() As Double, _
            ByRef Dvect() As Double, ByVal nNumRows As Integer, _
            ByRef bIsSingular As Boolean)
    Dim I As Integer, J As Integer, K As Integer
    Dim fScale As Double, Sigma As Double
    Dim fSum As Double, Tau As Double

    bIsSingular = False
    For K = 0 To (nNumRows - 1) - 1
      fScale = 0.0#
      For I = K To nNumRows - 1
        fScale = Max(fScale, Math.Abs(A(I, K)))
      Next I
      If fScale = 0.0# Then
        bIsSingular = True
        CVect(K) = 0
        Dvect(K) = 0.0#
      Else
        For I = K To nNumRows - 1
          A(I, K) = A(I, K) / fScale
        Next I
        fSum = 0
        I = K
        Do While I < nNumRows
          fSum = fSum + A(I, K) ^ 2
          I = I + 1
        Loop
        Sigma = Sign(Math.Sqrt(fSum), A(K, K))
        A(K, K) = A(K, K) + Sigma
        CVect(K) = Sigma * A(K, K)
        Dvect(K) = -fScale * Sigma
        For J = K + 1 To nNumRows - 1
          fSum = 0
          I = K
          Do While I < nNumRows
            fSum = fSum + A(I, K) * A(I, J)
            I = I + 1
          Loop
          Tau = fSum / CVect(K)
          For I = K To nNumRows - 1
            A(I, J) = A(I, J) - Tau * A(I, K)
          Next I
        Next J
      End If
    Next K
    Dvect((nNumRows - 1)) = A(nNumRows - 1, nNumRows - 1)
    If Dvect((nNumRows - 1)) = 0.0# Then
      bIsSingular = True
    End If
  End Sub


  Public Shared Sub QrSolve(ByRef A(,) As Double, ByRef CVect() As Double, _
         ByRef Dvect() As Double, ByRef BVect() As Double, _
         ByVal nNumRows As Integer)
    Dim I As Integer, J As Integer
    Dim fSum As Double, Tau As Double

    For J = 0 To (nNumRows - 1) - 1
      fSum = 0
      I = J
      Do While I < nNumRows
        fSum = fSum + A(I, J) * BVect(I)
        I = I + 1
      Loop
      Tau = fSum / CVect(J)
      For I = J To nNumRows - 1
        BVect(I) = BVect(I) - Tau * A(I, J)
      Next I
    Next J
    RSolve(A, Dvect, BVect, nNumRows)
  End Sub


  Public Shared Sub RSolve(ByRef A(,) As Double, ByRef d() As Double, ByRef B() As Double, ByVal nNumRows As Integer)
    Dim I As Integer, J As Integer
    Dim fSum As Double

    B(nNumRows - 1) = B(nNumRows - 1) / d(nNumRows - 1)
    For I = nNumRows - 2 To 0 Step -1
      fSum = 0
      J = I + 1
      Do While J < nNumRows
        fSum = fSum + A(I, J) * B(J)
        J = J + 1
      Loop
      B(I) = (B(I) - fSum) / d(I)
    Next I
  End Sub 'RSolve


  Public Shared Sub Rotate(ByRef R(,) As Double, ByRef qt(,) As Double, ByVal I As Integer, ByVal A As Double, ByVal B As Double, ByVal nNumRows As Integer)
    Dim J As Integer
    Dim C As Double, fact As Double, s As Double
    Dim W As Double, Y As Double

    If A = 0.0# Then
      C = 0.0#
      s = IIf(B >= 0.0#, 1.0#, -1.0#)
    ElseIf Math.Abs(A) > Math.Abs(B) Then
      fact = B / A
      C = Sign(1.0# / Math.Sqrt((1.0# + fact ^ 2)), A)
      s = fact * C
    Else
      fact = A / B
      s = Sign(1.0# / Math.Sqrt((1.0# + fact ^ 2)), B)
      C = fact * s
    End If
    For J = I To nNumRows - 1
      Y = R(I, J)
      W = R(I + 1, J)
      R(I, J) = C * Y - s * W
      R(I + 1, J) = s * Y + C * W
    Next J
    For J = 0 To nNumRows - 1
      Y = qt(I, J)
      W = qt(I + 1, J)
      qt(I, J) = C * Y - s * W
      qt(I + 1, J) = s * Y + C * W
    Next J
  End Sub 'Rotate

  Public Shared Sub QrUpdate(ByRef R(,) As Double, ByRef qt(,) As Double, ByRef U() As Double, ByRef V() As Double)
    Dim I As Integer, K As Integer
    Dim N As Integer

    N = UBound(U) + 1
    For K = N - 1 To 0 Step -1
      If U(K) <> 0.0# Then
        Exit For
      End If
    Next K
    If K < 0 Then
      K = 0
    End If
    For I = K - 1 To 0 Step -1
      Rotate(R, qt, I, U(I), -U((I + 1)), N)
      If U(I) = 0.0# Then
        U(I) = Math.Abs(U((I + 1)))
      ElseIf Math.Abs(U(I)) > Math.Abs(U((I + 1))) Then
        U(I) = Math.Abs(U(I)) * Math.Sqrt((1.0# + Square((U((I + 1)) / U(I)))))
      Else
        U(I) = Math.Abs(U((I + 1))) * Math.Sqrt((1.0# + Square((U(I) / U((I + 1))))))
      End If
    Next I
    For I = 0 To N - 1
      R(0, I) = R(0, I) + U(0) * V(I)
    Next I
    For I = 0 To K - 1
      Rotate(R, qt, I, R(I, I), -R(I + 1, I), N)
    Next I
  End Sub 'QrUpdate

  Public Shared Sub SolveQr(ByRef A(,) As Double, ByRef B() As Double, ByVal nCount As Integer)
    Dim bIsSingular As Boolean
    Dim nInitErrCount As Integer
    Dim C() As Double, d() As Double

    ReDim C(nCount)
    ReDim d(nCount)
    nInitErrCount = GetErrorCount()
    bIsSingular = False

    QrDecomp(A, C, d, nCount, bIsSingular)
    If GetErrorCount() > nInitErrCount Or bIsSingular Then
      Exit Sub
    End If
    QrSolve(A, C, d, B, nCount)
  End Sub 'SolveQr

  Public Shared Sub GaussJordan(ByRef A(,) As Double, ByRef B(,) As Double, ByVal nNumRows As Integer, ByVal nNumCols As Integer)
    Dim iCol As Integer
    Dim iRow As Integer
    Dim I As Integer, J As Integer, K As Integer
    Dim L As Integer, LL As Integer
    Dim big As Double, dum As Double, pivinv As Double
    Dim pivotIndexc() As Integer
    Dim pivotIndexr() As Integer
    Dim ipiv() As Integer

    iCol = 0
    iRow = 0
    On Error GoTo HandleErr
    ReDim pivotIndexc(nNumRows)
    ReDim pivotIndexr(nNumRows)
    ReDim ipiv(nNumRows)

    For J = 0 To nNumRows - 1
      ipiv(J) = 0
    Next J
    For I = 0 To nNumRows - 1
      big = 0.0#
      For J = 0 To nNumRows - 1
        If ipiv(J) <> 1 Then
          For K = 0 To nNumRows - 1
            If ipiv(K) = 0 Then
              If Math.Abs(A(J, K)) >= big Then
                big = Math.Abs(A(J, K))
                iRow = J
                iCol = K
              End If
            End If
          Next K
        End If
      Next J
      ipiv(iCol) = ipiv(iCol) + 1
      If iRow <> iCol Then
        For L = 0 To nNumRows - 1
          Swap(A(iRow, L), A(iCol, L))
        Next L
        For L = 0 To nNumCols - 1
          Swap(B(iRow, L), B(iCol, L))
        Next L
      End If
      pivotIndexr(I) = iRow
      pivotIndexc(I) = iCol
      If A(iCol, iCol) = 0.0# Then
        ReportError("Singular Matrix", "GaussJordan")
        Exit Sub
      End If
      pivinv = 1.0# / A(iCol, iCol)
      A(iCol, iCol) = 1.0#
      For L = 0 To nNumRows - 1
        A(iCol, L) = A(iCol, L) * pivinv
      Next L
      For L = 0 To nNumCols - 1
        B(iCol, L) = B(iCol, L) * pivinv
      Next L
      For LL = 0 To nNumRows - 1
        If LL <> iCol Then
          dum = A(LL, iCol)
          A(LL, iCol) = 0.0#
          For L = 0 To nNumRows - 1
            A(LL, L) = A(LL, L) - A(iCol, L) * dum
          Next L
          For L = 0 To nNumCols - 1
            B(LL, L) = B(LL, L) - B(iCol, L) * dum
          Next L
        End If
      Next LL
    Next I
    For L = nNumRows - 1 To 0 Step -1
      If pivotIndexr(L) <> pivotIndexc(L) Then
        For K = 0 To nNumRows - 1
          Swap(A(K, pivotIndexr(L)), A(K, pivotIndexc(L)))
        Next K
      End If
    Next L
ExitProc:
    Exit Sub

HandleErr:
    ReportError(Err.Description, "GaussJordan")
    Resume ExitProc
  End Sub

  Public Shared Sub SolveMultiRegLu(ByRef X(,) As Double, ByRef Y() As Double, ByVal nNumIndependentVars As Integer, ByRef B() As Double)
    Dim Xt(,) As Double, StatSumMat(,) As Double
    Dim nCount As Integer
    Dim pvtIdx() As Integer
    Dim Factor As Double
    Dim nInitErrCount As Integer

    nCount = nNumIndependentVars + 1
    ReDim pvtIdx(nCount - 1)
    ReDim B(nCount - 1)
    Factor = 1
    nInitErrCount = GetErrorCount()
    MatTrans(X, Xt)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultMat(Xt, X, StatSumMat)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultVect(Xt, Y, B)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuDecomp(StatSumMat, nCount, pvtIdx, Factor)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuBackSubst(StatSumMat, nCount, pvtIdx, B)
  End Sub 'SolveMultiRegLu


  Public Shared Sub SolveMultiRegLu_Ex(ByRef X(,) As Double, ByRef Y() As Double, ByVal nNumIndependentVars As Integer, ByRef B() As Double)
    Dim StatSumMatCopy(,) As Double
    Dim Xt(,) As Double, StatSumMat(,) As Double
    Dim BCopy() As Double
    Dim nCount As Integer
    Dim pvtIdx() As Integer
    Dim Factor As Double
    Dim nInitErrCount As Integer

    nCount = nNumIndependentVars + 1
    ReDim BCopy(1), StatSumMat(1, 1), StatSumMatCopy(1, 1)
    ReDim pvtIdx(nCount)
    BCopy(0) = 0
    Factor = 1
    nInitErrCount = GetErrorCount()
    MatTrans(X, Xt)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultMat(Xt, X, StatSumMat) ' or (X, Xt, ...)
    DuplicateMat(StatSumMat, StatSumMatCopy)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultVect(Xt, Y, B) ' was MatMultMat()
    DuplicateVector(B, BCopy)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuDecomp(StatSumMat, nCount, pvtIdx, Factor)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuImprove(StatSumMatCopy, StatSumMat, pvtIdx, BCopy, B, nCount)
  End Sub 'SolveMultiRegLu_Ex


  Public Shared Sub SolvePolyRegLu(ByRef X() As Double, ByRef Y() As Double, ByVal nOrder As Integer, ByRef B() As Double)
    Dim Xt(,) As Double, XData(,) As Double, StatSumMat(,) As Double
    Dim nCount As Integer
    Dim pvtIdx() As Integer
    Dim Factor As Double
    Dim I, J As Integer
    Dim nInitErrCount As Integer

    nCount = nOrder + 1
    nInitErrCount = GetErrorCount()
    Factor = 1
    ReDim XData(UBound(X), nCount - 1)
    ReDim pvtIdx(nCount - 1)

    For I = 0 To UBound(X) ' - 1
      For J = 0 To nCount - 1
        If J = 0 Then
          XData(I, J) = 1.0#
        ElseIf J = 1 Then
          XData(I, J) = X(I)
        Else
          XData(I, J) = X(I) ^ J
        End If
      Next J
    Next I

    MatTrans(XData, Xt)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultMat(Xt, XData, StatSumMat)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultVect(Xt, Y, B)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuDecomp(StatSumMat, nCount, pvtIdx, Factor)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuBackSubst(StatSumMat, nCount, pvtIdx, B)
  End Sub 'SolvePolyRegLu


  Public Shared Sub SolvePolyRegLu_Ex(ByRef X() As Double, ByRef Y() As Double, ByVal nOrder As Integer, ByRef B() As Double)
    Dim StatSumMatCopy(,) As Double
    Dim Xt(,) As Double, XData(,) As Double, StatSumMat(,) As Double
    Dim BCopy() As Double
    Dim nCount As Integer
    Dim pvtIdx() As Integer
    Dim Factor As Double
    Dim I As Integer, J As Integer
    Dim nInitErrCount As Integer


    nCount = nOrder + 1
    ReDim BCopy(1)
    BCopy(1) = 0
    nInitErrCount = GetErrorCount()
    Factor = 1
    ReDim XData(UBound(X), nCount - 1)
    'ReDim BCopy(1)
    ReDim pvtIdx(nCount - 1)

    For I = 0 To UBound(X)
      For J = 0 To nCount - 1
        If J = 0 Then
          XData(I, J) = 1.0#
        ElseIf J = 1 Then
          XData(I, J) = X(I)
        Else
          XData(I, J) = X(I) ^ J
        End If
      Next J
    Next I

    MatTrans(XData, Xt)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultMat(Xt, XData, StatSumMat)
    DuplicateMat(StatSumMat, StatSumMatCopy)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultVect(Xt, Y, B)
    DuplicateVector(B, BCopy)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuDecomp(StatSumMat, nCount, pvtIdx, Factor)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    LuImprove(StatSumMatCopy, StatSumMat, pvtIdx, BCopy, B, nCount)
  End Sub


  Public Shared Sub SolveMultiRegSvd(ByRef X(,) As Double, ByRef Y() As Double, ByVal nNumIndependentVars As Integer, ByRef B() As Double, ByRef Coeff() As Double)
    Dim Xt(,) As Double, StatSumMat(,) As Double
    Dim V(,) As Double
    Dim W() As Double
    Dim nInitErrCount As Integer
    Dim nCount As Integer

    nCount = nNumIndependentVars + 1
    ReDim W(nCount - 1)
    ReDim V(nCount, nCount - 1)
    nInitErrCount = GetErrorCount()
    MatTrans(X, Xt)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultMat(Xt, X, StatSumMat)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultVect(Xt, Y, B)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    SvdDecomp(StatSumMat, nCount, nCount, W, V)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    SvdBackSubst(StatSumMat, W, V, B, Coeff, nCount, nCount)
  End Sub


  Public Shared Sub SolvePolyRegSvd(ByRef X() As Double, ByRef Y() As Double, ByVal nOrder As Integer, ByRef B() As Double, ByRef Coeff() As Double)
    Dim Xt(,) As Double, XData(,) As Double, StatSumMat(,) As Double
    Dim V(,) As Double
    Dim W() As Double
    Dim nInitErrCount As Integer
    Dim nCount As Integer
    Dim I, J As Integer

    nCount = nOrder + 1
    ReDim W(nCount - 1)
    ReDim V(nCount, nCount - 1)
    nInitErrCount = GetErrorCount()
    ReDim XData(UBound(X), nCount - 1)

    For I = 0 To UBound(X)
      For J = 0 To nCount - 1
        If J = 0 Then
          XData(I, J) = 1.0#
        ElseIf J = 1 Then
          XData(I, J) = X(I)
        Else
          XData(I, J) = X(I) ^ J
        End If
      Next J
    Next I

    MatTrans(XData, Xt)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultMat(Xt, XData, StatSumMat)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultVect(Xt, Y, B)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    SvdDecomp(StatSumMat, nCount, nCount, W, V)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    SvdBackSubst(StatSumMat, W, V, B, Coeff, nCount, nCount)
  End Sub


  Public Shared Sub SolveMultiRegQr(ByRef X(,) As Double, ByRef Y() As Double, ByVal nNumIndependentVars As Integer, ByRef Coeff() As Double)
    Dim Xt(,) As Double, StatSumMat(,) As Double
    Dim C() As Double, d() As Double
    Dim bIsSingular As Boolean
    Dim nCount As Integer
    Dim nInitErrCount As Integer

    nCount = nNumIndependentVars + 1
    nInitErrCount = GetErrorCount()
    bIsSingular = False
    ReDim C(nCount - 1)
    ReDim d(nCount - 1)
    MatTrans(X, Xt)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultMat(Xt, X, StatSumMat)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultVect(Xt, Y, Coeff)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    QrDecomp(StatSumMat, C, d, nCount, bIsSingular)
    If GetErrorCount() > nInitErrCount Or bIsSingular Then
      Exit Sub
    End If
    QrSolve(StatSumMat, C, d, Coeff, nCount)
  End Sub 'SolveMultiRegQr


  Public Shared Sub SolvePolyRegQr(ByRef X() As Double, ByRef Y() As Double, ByVal nOrder As Integer, ByRef Coeff() As Double)
    Dim Xt(,) As Double, XData(,) As Double, StatSumMat(,) As Double
    Dim C() As Double, d() As Double
    Dim bIsSingular As Boolean
    Dim nCount As Integer
    Dim I As Integer, J As Integer
    Dim nInitErrCount As Integer

    nCount = nOrder + 1
    nInitErrCount = GetErrorCount()
    bIsSingular = False
    ReDim C(nCount - 1)
    ReDim d(nCount - 1)
    ReDim XData(UBound(X), nCount - 1)

    For I = 0 To UBound(X) ' - 1
      For J = 0 To nCount - 1
        If J = 0 Then
          XData(I, J) = 1.0#
        ElseIf J = 1 Then
          XData(I, J) = X(I)
        Else
          XData(I, J) = X(I) ^ J
        End If
      Next J
    Next I

    MatTrans(XData, Xt)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultMat(Xt, XData, StatSumMat)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    MatMultVect(Xt, Y, Coeff)
    If GetErrorCount() > nInitErrCount Then
      Exit Sub
    End If
    QrDecomp(StatSumMat, C, d, nCount, bIsSingular)
    If GetErrorCount() > nInitErrCount Or bIsSingular Then
      Exit Sub
    End If
    QrSolve(StatSumMat, C, d, Coeff, nCount)
  End Sub 'SolvePolyRegQr

  Public Shared Sub Initialize()
    ResetErrors() ' reset the error management data members
    mfBadDataVal = DEF_BAD_VAL ' set the default "BAD VALUE"
    ' initialize data for bit shifting functions
    InitShiftData()
  End Sub

End Class
